from ..LibStage import gen_multi_agent, QAStage, gen_agent
from .. import mock_client_once

"""
# 預設腳本 非等號比較
"""
bot_json = {
    "__MAIN_STAGES__": [
        {
            "stage_type": "__QA_STAGE__",
            "__STAGE_NAME__": "__FAQ階段__",
            "question": {
                "sys_welcome": "FAQ歡迎",
                "sys_refuse": "FAQ拒絕",
                "sys_complete": ""
            },
            "qa_threshold": 0.0,
            "corpus": {
                "推薦商品": "1"
            },
            "__SAVED_NAME__": {
                "__QA_RESPOND__": "FAQ_r1",
                "__QA_RESPOND_THRESHOLD__": "FAQ_th",
                "__QA_RESPOND_QUESTION__": "FAQ_q1",
                "__QA_RESPOND_SCORE__": "FAQ_s1",
                "__RUNNING_CORPUS__": "FAQ_c1"
            }
        },
        {
            "stage_type": "__LIB_SWITCH_STAGE__",
            "__STAGE_NAME__": "__FAQ切換階段__",
            "stages_filter": [
                [
                    ["FAQ_r1", "FAQ_s1"],
                    ["1", 0.7],
                    "_商品推薦1_",
                    ["=", ">"]
                ],
                [
                    ["FAQ_r1", "FAQ_s1"],
                    ["1", 0.5],
                    "_商品推薦2_",
                    ["=", "<"]
                ]
            ]
        }
    ], "_商品推薦1_": [
        {
            "stage_type": "__RE_STAGE__",
            "__STAGE_NAME__": "__商品推薦階段1__",
            "question": {
                "sys_welcome": "",
                "sys_refuse": "",
                "sys_complete": "推薦完成1"
            },
            "__DISABLE_WELCOME__": True
        }
    ], "_商品推薦2_": [
        {
            "stage_type": "__RE_STAGE__",
            "__STAGE_NAME__": "__商品推薦階段2__",
            "question": {
                "sys_welcome": "",
                "sys_refuse": "",
                "sys_complete": "推薦完成2"
            },
            "__DISABLE_WELCOME__": True
        }
    ]

}
data = {}
agent = gen_agent(bot_json)

#
text = "hi"
reply_text, data = mock_client_once(agent, text, data)
print(f"reply_text: {reply_text}")

text = "推薦商品"
reply_text, data = mock_client_once(agent, text, data)
print(f"reply_text: {reply_text}")


text = "hi"
reply_text, data = mock_client_once(agent, text, {})
# print(f"reply_text: {reply_text}")

text = "走開"
reply_text, data = mock_client_once(agent, text, data)
print(f"reply_text: {reply_text}")





from .__stage__ import Stage, __USER_TEXT__, StageType
from .__agent__ import Agent, MultiAgent
from .__tool__ import get_value_from_dict_by_multi_name, compute_by_string
from .jieba_zh import analyse
import re

__RE_STAGE__ = "__RE_STAGE__"
__QA_STAGE__ = "__QA_STAGE__"
__SWITCH_STAGE__ = StageType.SWITCH
__LIB_SWITCH_STAGE__ = "__LIB_SWITCH_STAGE__"

#
__QUESTIONS_LABEL__ = ["__SYS_QUESTION__", "question", "says"]
__WELCOME_SAYING_LABELS__ = ["__SYS_WELCOME__", "sys_reply_q1", "sys_welcome"]
__REFUSE_SAYING_LABELS__ = ["__SYS_REFUSE__", "sys_reply_q2", "sys_refuse"]
__COMPLETE_SAYING_LABELS__ = ["__SYS_COMPLETE__", "sys_reply_complete", "sys_complete"]
__DISABLE_WELCOME_LABEL__ = ["__DISABLE_WELCOME__", "__DISSABLE_Q1__", "DISSABLE_WELCOME"]
__DISABLE_REFUSE_LABEL__ = ["__DISABLE_REFUSE__"]

from .nlp_tool import similar


class QAStage(Stage):
    __QA_RESPOND__ = "__QA_RESPOND__"
    __QA_RESPOND_SCORE__ = "__QA_RESPOND_SCORE__"
    __QA_RESPOND_QUESTION__ = "__QA_RESPOND_QUESTION__"
    __QA_RESPOND_THRESHOLD__ = "__QA_RESPOND_THRESHOLD__"
    __QA_THRESHOLD__ = "qa_threshold"
    __IS_FITS__ = "is_fits"
    __SAVED_NAME__ = "__SAVED_NAME__"
    __CORPUS__ = "corpus"
    __QUESTIONS__ = "question"
    __SIMILAR_METHOD__ = "similar_method"
    __RUNNING_CORPUS__ = "__RUNNING_CORPUS__"
    __DISSABLE_Q1__ = "__DISSABLE_Q1__"
    __REFACTOR_QUESTION__ = "refactor_questions"

    def __init__(self, data):
        super(QAStage, self).__init__(**data)
        self.raw_data = data
        self.stage_type = data.get("stage_type")
        self.corpus = {}
        corpus = data.get(self.__CORPUS__)
        for k, v in corpus.items():
            self.corpus[k.replace(" ", "")] = v.replace(" ", "")

        #
        self.questions = get_value_from_dict_by_multi_name(d=data, names=__QUESTIONS_LABEL__)
        self.sys_reply_q1 = get_value_from_dict_by_multi_name(d=self.questions, names=__WELCOME_SAYING_LABELS__)
        self.sys_reply_q2 = get_value_from_dict_by_multi_name(d=self.questions, names=__REFUSE_SAYING_LABELS__)
        self.sys_reply_complete = get_value_from_dict_by_multi_name(d=self.questions,
                                                                    names=__COMPLETE_SAYING_LABELS__)
        self.saved_name: dict = data.get(self.__SAVED_NAME__, {})
        self.qa_threshold = data.get(self.__QA_THRESHOLD__, 0.1)

        # TODO: RENAME `refactor_questions` -> `refactor_corpus_questions`
        self.refactor_questions = data.get(self.__REFACTOR_QUESTION__, False)

        if get_value_from_dict_by_multi_name(d=data, names=__DISABLE_WELCOME_LABEL__, default=False):
            self.is_first_access = self.disabel_is_first_access

        self.disable_refuse_question = get_value_from_dict_by_multi_name(d=data, names=__DISABLE_REFUSE_LABEL__,
                                                                         default=False)

    def __request_similar_api__(self, text, corpus):
        return similar(text, corpus)

    def __encode_corpus__(self):
        source_corpus = list(self.corpus.keys())
        dict_corpus = {self.__keyword__(self.__stop_word__(s)): s for s in source_corpus}
        return list(self.corpus.keys() if not self.refactor_questions else dict_corpus.keys()), dict_corpus

    def __decode_corpus__(self, worker_response, new_dict_corpus):
        best_res_content = str(worker_response[0][0]).replace(" ", "")
        if self.refactor_questions:
            best_res_content = new_dict_corpus[best_res_content] if best_res_content in new_dict_corpus else '0'
        return best_res_content

    def is_fit_needs_n_gen_entity(self, kwargs) -> (bool, dict):
        user_text = kwargs.get(__USER_TEXT__, "")

        corpus, new_dict_corpus = self.__encode_corpus__()

        responds = self.__request_similar_api__(user_text, corpus)
        worker_response = responds["worker response"]["ans"]
        best_res_content = self.__decode_corpus__(worker_response, new_dict_corpus)

        ##
        best_res_score = worker_response[0][1]
        best_res_responds = self.corpus[best_res_content] if best_res_content in self.corpus else None

        ##
        __RUNNING_CORPUS__ = self.saved_name.get(self.__RUNNING_CORPUS__, self.__RUNNING_CORPUS__)
        __QA_RESPOND__ = self.saved_name.get(self.__QA_RESPOND__, self.__QA_RESPOND__)
        __QA_RESPOND_QUESTION__ = self.saved_name.get(self.__QA_RESPOND_QUESTION__, self.__QA_RESPOND_QUESTION__)
        __QA_RESPOND_SCORE__ = self.saved_name.get(self.__QA_RESPOND_SCORE__, self.__QA_RESPOND_SCORE__)
        __QA_RESPOND_THRESHOLD__ = self.saved_name.get(self.__QA_RESPOND_THRESHOLD__, self.__QA_RESPOND_THRESHOLD__)

        #
        kwargs = self.set_default_var(kwargs, __RUNNING_CORPUS__, corpus)
        kwargs = self.set_default_var(kwargs, __QA_RESPOND__, best_res_responds)
        kwargs = self.set_default_var(kwargs, __QA_RESPOND_QUESTION__, best_res_content)
        kwargs = self.set_default_var(kwargs, __QA_RESPOND_SCORE__, best_res_score)
        if best_res_score >= self.qa_threshold:
            pass_token = True
            kwargs = self.set_default_var(kwargs, __QA_RESPOND_THRESHOLD__, True)
        else:
            pass_token = False if self.disable_refuse_question is False else True

            kwargs = self.set_default_var(kwargs, __QA_RESPOND_THRESHOLD__, False)

        return pass_token, kwargs

    @staticmethod
    def disabel_is_first_access(kwargs, stage_id):
        return False

    @staticmethod
    def __stop_word__(text, pronouns=True, common_skip=True, road_name=True, symbol=True):
        text = text.replace(" ", "")
        filter_set = []
        if pronouns:
            filter_set += ["我", "你", "妳"]
        if common_skip:
            filter_set += ["請問", "不好意思", "這裡", "可以", "要", "怎麼", "如何"]
        if road_name:
            filter_set += ["路", "號"]
        if symbol:
            filter_set += ["-", "，", "？", "。"]
        for stopword in filter_set:
            text = text.replace(stopword, " ")
        return text

    @staticmethod
    def __keyword__(text, k=10):
        allowed_words = ""
        list_en = [i for i in "abcdefghijklmnopqrstuvwxyz"] + [i for i in "abcdefghijklmnopqrstuvwxyz".upper()]
        list_math = [str(i) for i in range(0, 10)]
        list_bus = ["紅", "棕", "綠", "橘", "藍", "黃", "區", "副"]
        new_text = analyse.tfidf(text, topK=k, withWeight=True)
        allowed_words += "".join([t for t, v in new_text] + list_bus + list_en + list_math)
        return_text = "".join([w for w in text if w in allowed_words])
        return return_text


class REStage(Stage):

    def __init__(self, data):
        super(REStage, self).__init__(**data)
        self.raw_data = data
        self.stage_type = data.get("stage_type")

        #
        self.questions = get_value_from_dict_by_multi_name(d=data, names=__QUESTIONS_LABEL__)
        self.sys_reply_q1 = get_value_from_dict_by_multi_name(d=self.questions, names=__WELCOME_SAYING_LABELS__)
        self.sys_reply_q2 = get_value_from_dict_by_multi_name(d=self.questions, names=__REFUSE_SAYING_LABELS__)
        self.sys_reply_complete = get_value_from_dict_by_multi_name(d=self.questions,
                                                                    names=__COMPLETE_SAYING_LABELS__)
        self.is_fits = data.get("is_fits", [])
        if get_value_from_dict_by_multi_name(d=data, names=__DISABLE_WELCOME_LABEL__, default=False):
            self.is_first_access = self.disabel_is_first_access

        self.disable_refuse_question = get_value_from_dict_by_multi_name(d=data, names=__DISABLE_REFUSE_LABEL__,
                                                                         default=False)

    @staticmethod
    def disabel_is_first_access(kwargs, stage_id):
        return False

    @staticmethod
    def __get_entity__(rule, text):
        entities = []
        results = re.findall(rule, text)
        for r in results:
            if isinstance(r, tuple):
                entities += list(r)
            else:
                entities.append(r)
        return entities

    def is_fit_needs_n_gen_entity(self, kwargs) -> (bool, dict):
        user_text = kwargs.get(__USER_TEXT__, "")
        pass_token = True

        missing_entities = []
        for (rule, entity_name) in self.is_fits:
            entities = self.__get_entity__(rule, user_text)
            if len(entities) > 0:
                kwargs = self.set_default_var(kwargs, entity_name, " ".join(entities))
            else:
                missing_entities.append(entity_name)

        for entity_name in missing_entities:
            if self.get_default_var(kwargs, entity_name) is None:
                pass_token = False

        if self.disable_refuse_question:
            pass_token = True

        return pass_token, kwargs


class LibSwitchStage(Stage):

    def __init__(self, data):
        super(LibSwitchStage, self).__init__(**data)
        self.raw_data = data
        self.stage_type = StageType.SWITCH
        self.stages_filter = data.get("stages_filter")

    @staticmethod
    def is_first_access(data, stage_id):
        return False

    def find_new_stages(self, kwargs):
        stages_searched = []
        for filter_tuple in self.stages_filter:

            if len(filter_tuple) == 3:
                label, text, stages_label = filter_tuple
                symbol = None
            elif len(filter_tuple) == 4:
                label, text, stages_label, symbol = filter_tuple
            else:
                raise SyntaxError("find_new_stages: filter_tuple format error")

            stages_searched.append(stages_label)
            #
            if label == "*" and text is True:
                return stages_label

            #
            assert not isinstance(label, list) or isinstance(label, type(
                text)), "切換路線之條件與限定結果的類型不一致，必須保持兩者都是`str` 或是兩者都是`list`"
            if not isinstance(label, list):
                label, text = [label], [text]

            assert len(label) == len(text), "切換路線之條件與限定結果的數量不一致"

            #
            if symbol is None:
                symbol = ["="] * int(len(text))
            elif not isinstance(symbol, list):
                symbol = [symbol]
            #
            assert len(symbol) == len(
                text), f"變數比較條件與限定結果的數量不一致len(symbol) == len(text): {symbol},{text}"

            #
            pass_token = True
            for l, s, t in zip(label, symbol, text):
                l_var = self.get_default_var(kwargs, l)
                pass_token = compute_by_string(l_var, s, t)

            if pass_token:
                return stages_label

        raise RuntimeError(f"Not Found Path: {stages_searched}")

    def is_fit_needs_n_gen_entity(self, kwargs) -> (bool, dict):
        raise RuntimeError


def gen_multi_agent(stage_dict: dict):
    stages = {}
    for stages_label, stage_jsons in stage_dict.items():
        stages[stages_label] = []
        for stage_json in stage_jsons:
            stage_class = __LIB_STAGES__[stage_json["stage_type"]]
            stages[stages_label].append(stage_class(stage_json))

    return MultiAgent(stages)


def gen_agent(stage_dict: dict):
    return gen_multi_agent(stage_dict)


#
__LIB_STAGES__ = {
    __RE_STAGE__: REStage,
    __QA_STAGE__: QAStage,
    __SWITCH_STAGE__: LibSwitchStage,
    __LIB_SWITCH_STAGE__: LibSwitchStage
}



import json
import uuid
import re

from .__agent__ import Agent
from .__stage__ import Stage, __USER_TEXT__, __SYS_REPLY__, StageStatus, \
    __LOCAL_VAR_LABEL__, __LOCAL_VAR_VALUE__

__MOCK_STAGES_LABEL_1__ = "__MOCK_STAGES_LABEL_1__"


def mock_client_with_test(agent, says, tests):
    data = {}
    for s, t in zip(says, tests):
        data[__USER_TEXT__] = s
        reply_text, data = agent.run_all_stages(**data)
        assert t == reply_text


def mock_client(agent, says, show_data=True, show_user_text=True):
    data = {}
    for s in says:
        data[__USER_TEXT__] = s
        reply_text, data = agent.run_all_stages(**data)
        if show_user_text:
            print("\t用戶:", s)
        print("系統:", reply_text)
        if show_data:
            print("\t系統資料:", data)


def mock_client_human(agent):
    data = {}
    while True:
        s = input("請輸入：")
        data[__USER_TEXT__] = s
        if s == "exit":
            break
        reply_text, data = agent.run_all_stages(**data)
        print("系統:", reply_text)


def mock_client_once(agent: Agent, text: str, data: dict):
    data[__USER_TEXT__] = text
    return agent.run_all_stages(**data)


def to_bot(agent: Agent, text: str, data: dict):
    return mock_client_once(agent, text, data)



import json
import uuid
import re

from .__stage__ import Stage, __USER_TEXT__, __SYS_REPLY__, StageStatus, \
    __LOCAL_VAR_LABEL__, __LOCAL_VAR_VALUE__, StageType, __PASS_TOKEN__, __KEEP_VAR__, \
    __KEEP_DEFAULT_VAR__


class Agent:

    def __init__(self, stages: [Stage]):
        self.stages = stages

    @staticmethod
    def is_not_complete(data) -> bool:
        status = Stage.get_sys_stage_status(data)
        return True if status in [StageStatus.FIRST, StageStatus.REFUSE] else False

    @staticmethod
    def is_final_stage(stages, idx) -> bool:
        return True if len(stages) == (idx + 1) else False

    @staticmethod
    def clear_sys_reply(data):
        data.pop(__SYS_REPLY__, None)
        return data

    @classmethod
    def get_sys_reply(cls, data):
        reply: list = data.get(__SYS_REPLY__, [])
        reply = [line for line in reply if line != ""]
        return reply

    @staticmethod
    def set_sys_reply(data, texts: [str]):
        data[__SYS_REPLY__] = texts
        return data

    def run_all_stages(self, **kwargs) -> (list, dict):
        #
        kwargs = self.clear_sys_reply(kwargs)

        #
        for idx, stage in enumerate(self.stages):
            # print(f"kwargs in: {kwargs}")
            kwargs = stage.run(**kwargs)
            if self.is_not_complete(kwargs) is True:
                # print(f"\tkwargs out: {kwargs}")
                return self.get_sys_reply(kwargs), kwargs
            else:
                if self.is_final_stage(self.stages, idx) is True:
                    # print(f"\tkwargs out: {kwargs}")
                    return self.get_sys_reply(kwargs), kwargs

        raise RuntimeError


class MultiAgent(Agent):
    __MAIN_STAGES__ = "__MAIN_STAGES__"
    __MAX_LEVEL__ = 10

    def __init__(self, stages: dict):
        assert self.__MAIN_STAGES__ in stages, "__MAIN_STAGES__ need in stages"
        super().__init__(stages[self.__MAIN_STAGES__])
        self.multi_stages = stages
        self.__STAGES_IDS__ = []

    def to_dict(self):
        re_dict = {}
        for key, stages in self.multi_stages.items():
            re_dict[key] = [s.raw_data for s in stages]
        return re_dict

    @staticmethod
    def is_normal_reply(result):
        for r in result:
            if not isinstance(r, str):
                return False
        return True

    @staticmethod
    def is_switch_stage(stage: Stage):
        return True if stage.stage_type == StageType.SWITCH else False

    def run_one_stages(self, stages, kwargs):
        for idx, stage in enumerate(stages):

            # clear data for loop map
            if stage.stage_id in self.__STAGES_IDS__:
                for rm_idx in self.__STAGES_IDS__[self.__STAGES_IDS__.index(stage.stage_id):]:

                    # clear pass-token
                    if rm_idx in kwargs[__PASS_TOKEN__]:
                        del kwargs[__PASS_TOKEN__][rm_idx]

                    # clear data of stage was deleted
                    if rm_idx in kwargs[__KEEP_VAR__]:
                        for var_name in kwargs[__KEEP_VAR__][rm_idx]:
                            if var_name in kwargs[__KEEP_VAR__][__KEEP_DEFAULT_VAR__]:
                                del kwargs[__KEEP_VAR__][__KEEP_DEFAULT_VAR__][var_name]
                        del kwargs[__KEEP_VAR__][rm_idx]

                # reset self.__STAGES_IDS__ because delete loop stage from user data
                self.__STAGES_IDS__ = self.__STAGES_IDS__[:self.__STAGES_IDS__.index(stage.stage_id)]

            #
            self.__STAGES_IDS__.append(stage.stage_id)

            ##
            if self.is_switch_stage(stage) is True:
                new_stages_label = stage.find_new_stages(kwargs)
                return self.multi_stages[new_stages_label], kwargs
            else:
                kwargs = stage.run(**kwargs)
                if self.is_not_complete(kwargs) is True:
                    return self.get_sys_reply(kwargs), kwargs
                else:
                    if self.is_final_stage(stages, idx) is True:
                        return self.get_sys_reply(kwargs), kwargs

    def run_all_stages(self, **kwargs) -> (list, dict):
        loop_level = 0
        self.__STAGES_IDS__ = []
        kwargs = self.clear_sys_reply(kwargs)
        result = self.stages
        while self.__MAX_LEVEL__ > loop_level:
            loop_level += 1
            #
            result, kwargs = self.run_one_stages(result, kwargs)
            #
            if self.is_normal_reply(result):
                return result, kwargs



from ..LibStage import gen_multi_agent, QAStage, gen_agent
from .. import mock_client_once

"""
# 預設腳本
"""
bot_json = {
    "__MAIN_STAGES__": [
        {
            "stage_type": "__RE_STAGE__",
            "__SYS_QUESTION__": {
                "__SYS_WELCOME__": ["歡迎句1","歡迎句2","歡迎句3"],
                "__SYS_REFUSE__": ["拒絕句1","拒絕句2","拒絕句3"],
                "__SYS_COMPLETE__": ["完成句1","完成句2","完成句3"]
            },
            "is_fits": [
                [
                    "(通過條件)+",
                    "set_level"
                ]
            ]
        }
    ]
        
}
data = {}
agent = gen_agent(bot_json)

#
text = "hi"
reply_text, data = mock_client_once(agent, text, data)
print(f"data: {data}")
assert ['歡迎句1'] == reply_text or ['歡迎句2'] == reply_text or ['歡迎句3'] == reply_text

text = "不要通過"
reply_text, data = mock_client_once(agent, text, data)
print(f"data: {data}")
assert ['拒絕句1'] == reply_text or ['拒絕句2'] == reply_text or ['拒絕句3'] == reply_text

text = "通過條件"
reply_text, data = mock_client_once(agent, text, data)
print(f"data: {data}")
assert ['完成句1'] == reply_text or ['完成句2'] == reply_text or ['完成句3'] == reply_text



from ..LibStage import gen_multi_agent, QAStage
from .. import mock_client_once,to_bot

bot_json = {
    "__MAIN_STAGES__": [
        {
            "stage_type": "__RE_STAGE__",
            "question": {
                "sys_reply_q1": "請問是要做哪種票種呢？",
                "sys_reply_q2": "請說『月票』或是『單程票』",
                "sys_reply_complete": "好的，將開始訂購 %%set_level%% "
            },
            "is_fits": [
                [
                    "(月票|1280|長期票|定期票)+",
                    "set_level"
                ],
                [
                    "(單程票|單程|一次)+",
                    "set_level"
                ]
            ]
        },
        {
            "stage_type": "Switch",
            "stages_filter": [
                [
                    "set_level",
                    "月票",
                    "_月票_"
                ],
                [
                    "set_level",
                    "1280",
                    "_月票_"
                ],
                [
                    "set_level",
                    "長期票",
                    "_月票_"
                ],
                [
                    "set_level",
                    "定期票",
                    "_月票_"
                ],
                [
                    "set_level",
                    "單程票",
                    "_單程票_"
                ],
                [
                    "set_level",
                    "單程",
                    "_單程票_"
                ],
                [
                    "set_level",
                    "一次",
                    "_單程票_"
                ]
            ]
        }
    ],
    "_月票_": [
        {
            "stage_type": "__RE_STAGE__",
            "question": {
                "sys_reply_q1": "月票的價格為 1280元，是否確認？",
                "sys_reply_q2": "月票的價格為 1280元，是否確認？請回答『是』或『否』",
                "sys_reply_complete": "好的，確認您使用 %%set_level%% 車廂的意願為 『 %%user_status%% 』，\n            感謝您的使用。\n        "
            },
            "is_fits": [
                [
                    "(是|好的|好|沒問題)+$",
                    "user_status"
                ],
                [
                    "(否|不|不行|不要|不好)+$",
                    "user_status"
                ]
            ]
        }
    ],
    "_單程票_": [
        {
            "stage_type": "__RE_STAGE__",
            "question": {
                "sys_reply_q1": "",
                "sys_reply_q2": "",
                "sys_reply_complete": "如果要訂購單程票，請使用票卷機，感謝您的使用。\n        "
            },
            "is_fits": [],
            "__DISSABLE_Q1__": True
        }
    ]
}
data = {}
agent = gen_multi_agent(bot_json)

#
text = "hi"
reply_text, data = mock_client_once(agent, text, data)
print(f"data: {data}")
print(f"reply_text: {reply_text}")

#
text = "月票"
reply_text, data = mock_client_once(agent, text, data)
print(f"data: {data}")
print(f"reply_text: {reply_text}")

#
text = "好"
reply_text, data = mock_client_once(agent, text, data)
print(f"data: {data}")
print(f"reply_text: {reply_text}")

'''
#
#
#
#
#
#
#
#
#
#
#
#
#
#
'''

bot_json = {
    "__MAIN_STAGES__": [
        {
            "stage_type": "__RE_STAGE__",
            "question": {
                "sys_reply_q1": "哈囉請問要做什麼？ 目前提供『問答』和『訂票』服務",
                "sys_reply_q2": "目前只提供『問答』和『訂票』服務喔",
                "sys_reply_complete": "好的，將開始 『 %%selected_service%% 』 "
            },
            "is_fits": [
                [
                    "(問答|問題|詢問)+",
                    "selected_service"
                ],
                [
                    "(訂票|票價|買票)+",
                    "selected_service"
                ]
            ]
        },
        {
            "stage_type": "Switch",
            "stages_filter": [
                [
                    "selected_service",
                    "訂票",
                    "_訂票_"
                ],
                [
                    "selected_service",
                    "買票",
                    "_訂票_"
                ],
                [
                    "selected_service",
                    "票價",
                    "_訂票_"
                ],
                [
                    "selected_service",
                    "問答",
                    "_問答_"
                ],
                [
                    "selected_service",
                    "問題",
                    "_問答_"
                ],
                [
                    "selected_service",
                    "詢問",
                    "_問答_"
                ]
            ]
        }
    ],
    "_問答_": [
        {
            "stage_type": "__QA_STAGE__",
            "corpus": {
                "廁所在哪裡": "這裡沒有廁所",
                "詢問處在哪裡": "這裡沒有詢問處",
                "診所在哪裡": "這裡沒有診所",
            },
            "question": {
                "sys_reply_q1": "請問有什麼問題呢？",
                "sys_reply_q2": "",
                "sys_reply_complete": "我有 %%__QA_RESPOND_SCORE__%% 的信心覺得您要問：<br> \n        %%__QA_RESPOND_QUESTION__%% <br> \n        答案是 %%__QA_RESPOND__%% "
            },
            "is_fits": []
        }
    ],
    "_訂票_": [
        {
            "stage_type": "__RE_STAGE__",
            "question": {
                "sys_reply_q1": "請問是要做哪種票種呢？",
                "sys_reply_q2": "請說『月票』或是『單程票』",
                "sys_reply_complete": "好的，將開始訂購 %%set_level%% "
            },
            "is_fits": [
                [
                    "(月票|1280|長期票|定期票)+",
                    "set_level"
                ],
                [
                    "(單程票|單程|一次)+",
                    "set_level"
                ]
            ]
        },
        {
            "stage_type": "Switch",
            "stages_filter": [
                [
                    "set_level",
                    "月票",
                    "_月票_"
                ],
                [
                    "set_level",
                    "1280",
                    "_月票_"
                ],
                [
                    "set_level",
                    "長期票",
                    "_月票_"
                ],
                [
                    "set_level",
                    "定期票",
                    "_月票_"
                ],
                [
                    "set_level",
                    "單程票",
                    "_單程票_"
                ],
                [
                    "set_level",
                    "單程",
                    "_單程票_"
                ],
                [
                    "set_level",
                    "一次",
                    "_單程票_"
                ]
            ]
        }
    ],
    "_月票_": [
        {
            "stage_type": "__RE_STAGE__",
            "question": {
                "sys_reply_q1": "月票的價格為 1280元，是否確認？",
                "sys_reply_q2": "月票的價格為 1280元，是否確認？請回答『是』或『否』",
                "sys_reply_complete": "好的，確認您使用 %%set_level%% 車廂的意願為 『 %%user_status%% 』，\n            感謝您的使用。\n        "
            },
            "is_fits": [
                [
                    "(是|好的|好|沒問題)+$",
                    "user_status"
                ],
                [
                    "(否|不|不行|不要|不好)+$",
                    "user_status"
                ]
            ]
        }
    ],
    "_單程票_": [
        {
            "stage_type": "__RE_STAGE__",
            "question": {
                "sys_reply_q1": "",
                "sys_reply_q2": "",
                "sys_reply_complete": "如果要訂購單程票，請使用票卷機，感謝您的使用。\n        "
            },
            "is_fits": [],
            "__DISSABLE_Q1__": True
        }
    ]
}
data = {}
agent = gen_multi_agent(bot_json)

#
text = "hi"
reply_text, data = mock_client_once(agent, text, data)
print(f"data: {data}")
print(f"reply_text: {reply_text}")

#
text = "我要月票"
reply_text, data = mock_client_once(agent, text, data)
print(f"data: {data}")
print(f"reply_text: {reply_text}")

#
text = "好"
reply_text, data = mock_client_once(agent, text, data)
print(f"data: {data}")
print(f"reply_text: {reply_text}")

'''
#
#
#
#
#
#
#
#
#
#
#
#
#
#
'''

from ..LibStage import __LIB_STAGES__, QAStage
import requests

"""
##########
__NEW QAWorkerSTAGE__
##########
"""
__NEW_QUESTIONANSWER__ = "NEW_QUESTIONANSWER"


class QAWorkerSTAGE(QAStage):
    def __init__(self, data):
        super(QAWorkerSTAGE, self).__init__(data)
        self.similar_method = data.get(self.__SIMILAR_METHOD__, "worker_api")
        self.__NLPCORESERVER__ = "http://52.147.71.0:8000"

    def __request_similar_api__(self, text, corpus):
        res = requests.post(url=f"{self.__NLPCORESERVER__}/jobs/{self.similar_method}", json={
            "sentence": [
                text
            ],
            "corpus": corpus})
        return res.json()


# 增加自訂義的類別
__LIB_STAGES__[__NEW_QUESTIONANSWER__] = QAWorkerSTAGE

"""
##########
Bot
##########
"""
bot_json = {
    "__MAIN_STAGES__": [
        {
            "stage_type": "__RE_STAGE__",
            "question": {
                "sys_reply_q1": "哈囉請問要做什麼？ 目前提供『問答』和『訂票』服務",
                "sys_reply_q2": "目前只提供『問答』和『訂票』服務喔",
                "sys_reply_complete": "好的，將開始 『 %%selected_service%% 』 "
            },
            "is_fits": [
                [
                    "(問答|問題|詢問)+",
                    "selected_service"
                ],
                [
                    "(訂票|票價|買票)+",
                    "selected_service"
                ]
            ]
        },
        {
            "stage_type": "Switch",
            "stages_filter": [
                [
                    "selected_service",
                    "訂票",
                    "_訂票_",
                    "="
                ],
                [
                    "selected_service",
                    "買票",
                    "_訂票_",
                    "="
                ],
                [
                    "selected_service",
                    "票價",
                    "_訂票_",
                    "="
                ],
                [
                    "selected_service",
                    "問答",
                    "_問答_", "="

                ],
                [
                    "selected_service",
                    "問題",
                    "_問答_", "="
                ],
                [
                    "selected_service",
                    "詢問",
                    "_問答_", "="
                ]
            ]
        }
    ],
    "_死亡_": [
        {
            "stage_type": "__RE_STAGE__",
            "question": {
                "sys_reply_q1": "",
                "sys_reply_q2": "",
                "sys_reply_complete": "死亡路線 "
            },
            "DISSABLE_WELCOME": True
        }

    ],
    "_存活_": [
        {
            "stage_type": "__RE_STAGE__",
            "question": {
                "sys_reply_q1": "存活路線 1",
                "sys_reply_q2": "存活路線 2",
                "sys_reply_complete": "存活路線 "
            },
            "DISSABLE_WELCOME": True
        }

    ],
    "_問答_": [
        {
            "stage_type": __NEW_QUESTIONANSWER__,
            "corpus": {
                "廁所在哪裡": "這裡沒有廁所",
                "詢問處在哪裡": "這裡沒有詢問處",
                "診所在哪裡": "這裡沒有診所",
            },
            "question": {
                "sys_reply_q1": "請問有什麼問題呢？",
                "sys_reply_q2": "",
                "sys_reply_complete": "我有 %%__QA_RESPOND_SCORE__%% 的信心覺得您要問：<br> \n        %%__QA_RESPOND_QUESTION__%% <br> \n        答案是 %%__QA_RESPOND__%% "
            },
            "is_fits": []
        }, {
            "stage_type": "Switch",
            "stages_filter": [
                [
                    "__QA_RESPOND_SCORE__", 0.5, "_存活_", ">"
                ],
                [
                    "__QA_RESPOND_SCORE__", 0.5, "_死亡_", "<"
                ]
            ]
        }

    ],
    "_訂票_": [
        {
            "stage_type": "__RE_STAGE__",
            "question": {
                "sys_reply_q1": "請問是要做哪種票種呢？",
                "sys_reply_q2": "請說『月票』或是『單程票』",
                "sys_reply_complete": "好的，將開始訂購 %%set_level%% "
            },
            "is_fits": [
                [
                    "(月票|1280|長期票|定期票)+",
                    "set_level"
                ],
                [
                    "(單程票|單程|一次)+",
                    "set_level"
                ]
            ]
        },
        {
            "stage_type": "Switch",
            "stages_filter": [
                [
                    "set_level",
                    "月票",
                    "_月票_"
                ],
                [
                    "set_level",
                    "1280",
                    "_月票_"
                ],
                [
                    "set_level",
                    "長期票",
                    "_月票_"
                ],
                [
                    "set_level",
                    "定期票",
                    "_月票_"
                ],
                [
                    "set_level",
                    "單程票",
                    "_單程票_"
                ],
                [
                    "set_level",
                    "單程",
                    "_單程票_"
                ],
                [
                    "set_level",
                    "一次",
                    "_單程票_"
                ]
            ]
        }
    ],
    "_月票_": [
        {
            "stage_type": "__RE_STAGE__",
            "question": {
                "sys_reply_q1": "月票的價格為 1280元，是否確認？",
                "sys_reply_q2": "月票的價格為 1280元，是否確認？請回答『是』或『否』",
                "sys_reply_complete": "好的，確認您使用 %%set_level%% 車廂的意願為 『 %%user_status%% 』，\n            感謝您的使用。\n        "
            },
            "is_fits": [
                [
                    "(是|好的|好|沒問題)+$",
                    "user_status"
                ],
                [
                    "(否|不|不行|不要|不好)+$",
                    "user_status"
                ]
            ]
        }
    ],
    "_單程票_": [
        {
            "stage_type": "__RE_STAGE__",
            "question": {
                "sys_reply_q1": "",
                "sys_reply_q2": "",
                "sys_reply_complete": "如果要訂購單程票，請使用票卷機，感謝您的使用。\n        "
            },
            "is_fits": [],
            "__DISSABLE_Q1__": True
        }
    ]
}
data = {}
agent = gen_multi_agent(bot_json)

#
text = "hi"
reply_text, data = to_bot(agent, text, data)
print(f"data: {data}")
print(f"reply_text: {reply_text}")

#
text = "我有點問題"
reply_text, data = to_bot(agent, text, data)
print(f"data: {data}")
print(f"reply_text: {reply_text}")

#
text = "附近有廁所嗎"
reply_text, data = to_bot(agent, text, data)
print(f"data: {data}")
print(f"reply_text: {reply_text}")




from . import Stage, Agent, mock_client, mock_client_human, __USER_TEXT__
import re


__HELLO_STAGE__ = "HELLO_STAGE"
__CLEAR_STAGE__ = "CLEAR_STAGE"
__ORDER_HELLO_STAGE__ = "ORDER_HELLO_STAGE"
__ORDER_START_STAGE__ = "__ORDER_START_STAGE__"
__ORDER_OUT_STAGE__ = "__ORDER_OUT_STAGE__"
__ORDER_CONFIRM_STAGE__ = "__ORDER_CONFIRM_STAGE__"


def fit_re(rule, text) -> bool:
    return False if re.match(rule, text) is None else True      # 不在起始位置匹配


class CleanStage(Stage):

    def __init__(self):
        super(CleanStage, self).__init__()
        self.stage_type = __CLEAR_STAGE__
        self.sys_reply_q1 = "您好目前您已經完成訂票了，如果想重新體驗，請輸入『我要重置』"
        self.sys_reply_q1 = f"""
            好的，您已經完成訂票了
            您預計將在
            %%{OrderStartStage.ON_BOARD}%% 上車
            在
             {self.get_default_var_label(OrderEndStage.OUT_BOARD)} 下車
            如果想重新體驗，請輸入『我要重置』
        """
        self.sys_reply_q2 = "如果想重新體驗，請輸入『我要重置』，想結束請輸入exit"
        self.sys_reply_complete = "已經重置"

    def is_fit_needs_n_gen_entity(self, kwargs) -> (bool, dict):
        if fit_re(".*重置.*", kwargs.get(__USER_TEXT__, "")):
            return True, None
        else:
            return False, kwargs


class HelloStage(Stage):

    def __init__(self):
        super(HelloStage, self).__init__()
        self.stage_type = __HELLO_STAGE__
        self.sys_reply_q1 = "歡迎光臨～"
        self.sys_reply_q2 = "歡迎跟我說一聲hi喔"
        self.sys_reply_complete = "哈囉您好，請問是來訂票的嗎？"

    @staticmethod
    def is_fit_needs_n_gen_entity(kwargs) -> (bool,dict):
        if fit_re(".*hi.*", kwargs[__USER_TEXT__]):
            return True, kwargs
        else:
            return False, kwargs


class OrderHelloStage(Stage):

    def __init__(self):
        super(OrderHelloStage, self).__init__()
        self.stage_type = __ORDER_HELLO_STAGE__
        self.sys_reply_q2 = "如果想訂票，請說 「線上訂票」，或是「我要訂票」"
        self.sys_reply_complete = "接下來進行訂票流程"

    @staticmethod
    def is_first_access(data, stage_id):
        return False

    @staticmethod
    def is_fit_needs_n_gen_entity(kwargs) -> (bool, dict):
        if fit_re(".*訂票.*", kwargs.get(__USER_TEXT__, "")):
            return True, kwargs
        else:
            return False, kwargs


class OrderStartStage(Stage):
    ON_BOARD = "on_board_station"

    def __init__(self):
        super(OrderStartStage, self).__init__()
        self.stage_type = __ORDER_START_STAGE__
        self.sys_reply_q1 = "請問是哪一站上車呢？"
        self.sys_reply_q2 = "請說站名，例如『板橋火車站』"
        self.sys_reply_complete = f"好的，您將從 {self.get_default_var_label(self.ON_BOARD)} 出發"

    def is_fit_needs_n_gen_entity(self, kwargs) -> (bool, dict):
        user_text = kwargs.get(__USER_TEXT__, "")
        if fit_re(".*車站.*", user_text):
            kwargs = self.set_default_var(kwargs, self.ON_BOARD, user_text)
            return True, kwargs
        else:
            return False, kwargs


class OrderEndStage(Stage):
    OUT_BOARD = "out_station"

    def __init__(self):
        super(OrderEndStage, self).__init__()
        self.stage_type = __ORDER_OUT_STAGE__
        self.sys_reply_q1 = "請問是哪一站下車呢？"
        self.sys_reply_q2 = "請說站名，例如『板橋火車站』"
        self.sys_reply_complete = f"好的，您將從 {self.get_default_var_label(self.OUT_BOARD)} 下車"

    def is_fit_needs_n_gen_entity(self, kwargs) -> (bool, dict):
        user_text = kwargs.get(__USER_TEXT__, "")
        if fit_re(".*車站.*", user_text):
            kwargs = self.set_default_var(kwargs, self.OUT_BOARD, user_text)
            return True, kwargs
        else:
            return False, kwargs


class OrderConfirmStage(Stage):

    def __init__(self):
        super(OrderConfirmStage, self).__init__()
        self.stage_type = __ORDER_CONFIRM_STAGE__
        self.sys_reply_complete = f"好的，您完成訂票了\n {self.get_default_var_label(OrderStartStage.ON_BOARD)} 上車\n {self.get_default_var_label(OrderEndStage.OUT_BOARD)} 下車"

    @staticmethod
    def is_first_access(data, stage_id):
        return False

    def is_fit_needs_n_gen_entity(self, kwargs) -> (bool, dict):
        return True, kwargs






import json
import uuid
import re
import random
__USER_TEXT__ = "USER_TEXT"
__PASS_TOKEN__ = "PASS_TOKEN"
__SYS_REPLY__ = "SYS_REPLY"
__SYS_STAGE__ = "SYS_STAGE"
__KEEP_VAR__ = "KEEP_VAR"
__KEEP_DEFAULT_VAR__ = "KEEP_DEFAULT_VAR"
__LOCAL_VAR_LABEL__ = "__LOCAL_VAR_LABEL__"
__LOCAL_VAR_VALUE__ = "__LOCAL_VAR_VALUE__"


class StageStatus:
    INIT = "INIT"
    FIRST = "FIRST"
    REFUSE = "REFUSE"
    COMPLETE = "COMPLETE"


class StageType:
    BASE = "BASE"
    SWITCH = "Switch"


class Stage:
    def __init__(self, **kwargs):
        self.stage_type = StageType.BASE
        self.stage_uuid_name = kwargs.get("__STAGE_NAME__", str(uuid.uuid4())[:15])
        self.stage_id = f"{self.stage_type}:{self.stage_uuid_name}"
        self.data = kwargs
        self.sys_reply_q1 = "init sys reply"  # init
        self.sys_reply_q2 = "refuse sys reply"  # refuse
        self.sys_reply_complete = "complete sys reply"  # complete

    @staticmethod
    def set_sys_stage_status(data: dict, label: str) -> dict:
        data.update({__SYS_STAGE__: label})
        return data

    @staticmethod
    def get_sys_stage_status(data: dict) -> str:
        return data.get(__SYS_STAGE__, None)

    @staticmethod
    def is_first_access(data, stage_id):
        # First Access
        if __PASS_TOKEN__ not in data:
            return True
        if stage_id not in data[__PASS_TOKEN__]:
            return True
        return False

    @staticmethod
    def clear_user_text(data):
        # First Access
        _ = data.pop(__USER_TEXT__, None)
        return data

    @classmethod
    def save_user_text(cls, data, stage_id, label):
        user_text = cls.get_user_text(data)
        return cls.set_var(data, stage_id, label, user_text)

    @staticmethod
    def get_user_text(data):
        return data.get(__USER_TEXT__, None)

    @staticmethod
    def is_token_pass(data: dict, stage_id) -> bool:
        token = data.get(__PASS_TOKEN__, None)
        if token is None:
            return False
        if isinstance(token, dict) and token.get(stage_id, None) is True:
            return True
        else:
            return False

    @staticmethod
    def set_none_token_pass(data, stage_id):
        if __PASS_TOKEN__ not in data:
            data[__PASS_TOKEN__] = {}
        data[__PASS_TOKEN__][stage_id] = None
        return data

    @staticmethod
    def set_true_token_pass(data, stage_id):
        if __PASS_TOKEN__ not in data:
            data[__PASS_TOKEN__] = {}
        data[__PASS_TOKEN__][stage_id] = True
        return data

    def is_fit_needs_n_gen_entity(self, kwargs) -> (bool, dict):
        kwargs = self.set_default_var(kwargs, __LOCAL_VAR_LABEL__, __LOCAL_VAR_VALUE__)
        # if want to reset data, return (True,None) to replace to (True,kwargs)
        return True, kwargs

    @staticmethod
    def is_var_label(label_string):
        return None if re.match("^var\t.*\t.*\tvar$", label_string) is None else label_string.split("\t")

    @staticmethod
    def is_var_label_human(label_string):
        return None if re.match("^%%.*%%$", label_string) is None else label_string[2:-2]

    @classmethod
    def get_default_var_label(cls, label: str):
        return cls.get_var_label(__KEEP_DEFAULT_VAR__, label)

    def set_default_var(self, data, label, save_text):
        data = self.set_var(data, self.stage_id, label, save_text)
        return self.set_var(data, __KEEP_DEFAULT_VAR__, label, save_text)

    def get_default_var(self, data, label):
        return self.get_var(data, __KEEP_DEFAULT_VAR__, label)

    @staticmethod
    def get_var_label(stage_id: str, label: str):
        assert "\t" not in stage_id
        assert "\t" not in label
        return f"var\t{stage_id}\t{label}\tvar"

    @classmethod
    def set_var(cls, data, stage_id, label, save_text):
        if __KEEP_VAR__ not in data:
            data[__KEEP_VAR__] = {}
        if stage_id not in data[__KEEP_VAR__]:
            data[__KEEP_VAR__][stage_id] = {}

        data[__KEEP_VAR__][stage_id][label] = save_text
        return data

    @classmethod
    def get_var(cls, data, stage_id, label):
        try:
            return data[__KEEP_VAR__][stage_id][label]
        except KeyError:
            return None

    @staticmethod
    def set_sys_reply(data, sys_reply_text):
        if __SYS_REPLY__ in data:
            data[__SYS_REPLY__].append(sys_reply_text)
        else:
            data[__SYS_REPLY__] = [sys_reply_text]
        return data

    @classmethod
    def replace_var_ticket_to_string(cls, kwargs, sys_reply):
        # print(f"kwargs: {kwargs}")
        # print(f"sys_reply: {sys_reply}")
        __NEXT_LINE__ = "||n||"
        # insert ENTITY
        sys_reply_complete_refactor = ""
        ##
        sys_reply = sys_reply.replace("\n", __NEXT_LINE__)

        #
        for var in sys_reply.split(" "):

            # parse
            var_ticket_human = cls.is_var_label_human(var)
            if var_ticket_human is not None:
                var_ticket = cls.get_default_var_label(var_ticket_human).split("\t")
            else:
                var_ticket = cls.is_var_label(var)

            # replace
            if var_ticket is not None:
                try:
                    var_value = kwargs[__KEEP_VAR__][var_ticket[1]][var_ticket[2]]
                    if isinstance(var_value, str):
                        sys_reply_complete_refactor += var_value
                    elif isinstance(var_value, float):
                        sys_reply_complete_refactor += str(round(var_value, 4))
                    else:
                        sys_reply_complete_refactor += str(var_value)

                except KeyError:
                    sys_reply_complete_refactor += var
            else:
                var = var.replace(__NEXT_LINE__, "\n")
                sys_reply_complete_refactor += var
        return sys_reply_complete_refactor

    def run(self, **kwargs):
        if self.is_token_pass(data=kwargs, stage_id=self.stage_id) is True:
            kwargs = self.set_sys_stage_status(kwargs, StageStatus.COMPLETE)
            return kwargs

        if self.is_first_access(kwargs, self.stage_id) is True:
            kwargs = self.set_none_token_pass(data=kwargs, stage_id=self.stage_id)
            kwargs = self.set_sys_stage_status(kwargs, StageStatus.FIRST)
            sys_reply = self.__choice_a_reply__(self.sys_reply_q1)
            #
            sys_reply = self.replace_var_ticket_to_string(kwargs, sys_reply)
            return self.set_sys_reply(kwargs, sys_reply)

        is_fit_token, kwargs = self.is_fit_needs_n_gen_entity(kwargs)
        if is_fit_token is False:  # q1
            kwargs = self.set_sys_stage_status(kwargs, StageStatus.REFUSE)
            sys_reply = self.__choice_a_reply__(self.sys_reply_q2)

            sys_reply = self.replace_var_ticket_to_string(kwargs, sys_reply)
            return self.set_sys_reply(kwargs, sys_reply)

        elif is_fit_token is True and kwargs is not None:  # q2
            kwargs = self.save_user_text(kwargs, self.stage_id, __USER_TEXT__)
            kwargs = self.clear_user_text(kwargs)
            kwargs = self.set_true_token_pass(kwargs, stage_id=self.stage_id)
            kwargs = self.set_sys_stage_status(kwargs, StageStatus.COMPLETE)

            # insert ENTITY
            sys_reply = self.replace_var_ticket_to_string(kwargs, self.__choice_a_reply__(self.sys_reply_complete))
            return self.set_sys_reply(kwargs, sys_reply)

        elif is_fit_token is True and kwargs is None:  # reset route
            sys_reply = self.replace_var_ticket_to_string(kwargs, self.__choice_a_reply__(self.sys_reply_complete))
            return self.set_sys_reply({}, sys_reply)

    @staticmethod
    def __choice_a_reply__(sys_reply):
        if isinstance(sys_reply, list):
            sys_reply = random.choice(sys_reply)
            return sys_reply
        return sys_reply


a = Stage()
res = a.run()
assert res.get(__SYS_STAGE__, None) == StageStatus.FIRST
res[__USER_TEXT__] = "hi"
res.pop(__SYS_REPLY__)
res = a.run(**res)
assert res.get(__SYS_STAGE__, None) == StageStatus.COMPLETE
res.pop(__SYS_REPLY__)
res = a.run(**res)
assert res.get(__SYS_STAGE__, None) == StageStatus.COMPLETE


def result_format(sentence,corpus,ans,score):
    return {
      "get parameter": {
        "sentence": [
          sentence
        ],
        "corpus": corpus
      },
      "worker response": {
        "ans": [
          [
            ans,
            score
          ]

        ]
      }
    }

def MED(sent_01, sent_02):
    n = len(sent_01)
    m = len(sent_02)

    matrix = [[i + j for j in range(m + 1)] for i in range(n + 1)]

    for i in range(1, n + 1):
        for j in range(1, m + 1):
            if sent_01[i - 1] == sent_02[j - 1]:
                d = 0
            else:
                d = 1

            matrix[i][j] = min(matrix[i - 1][j] + 1, matrix[i][j - 1] + 1, matrix[i - 1][j - 1] + d)

    distance_score = matrix[n][m]

    return distance_score


def similar(sentence, corpus):
    ans = None
    min_distance = 999
    max_distance = 1
    for c in corpus:
        distance = MED(sentence, c)
        max_distance = distance if distance > max_distance else max_distance
        if distance < min_distance:
            min_distance = distance
            ans = c
    score = 1-(min_distance/max_distance)
    result = result_format(sentence, corpus, ans, score)
    return result
#
# corpus= [
#     "我想去醫院",
#     "我想去大學",
#     "我想去台北",
#     "我想去跑步",
#     "我想去公園"
# ]
# sentence="我想去大公園"
# print(similar(sentence,corpus))


def get_value_from_dict_by_multi_name(d: dict, names: [str], default=None):
    for name in names:
        if name in d:
            return d[name]
    return default


def compute_by_string(a, symbol, b):
    pass_token = True
    if symbol != "=":
        a, b = float(a), float(b)
    if symbol == "=":
        if a != b:
            pass_token = False
    elif symbol == ">":
        if not a > b:
            pass_token = False
    elif symbol == "<":
        if not a < b:
            pass_token = False
    elif symbol == "<=":
        if not a < b:
            pass_token = False
    elif symbol == ">=":
        if not a >= b:
            pass_token = False
    return pass_token


import setuptools
with open("./ConversationAgent/README.md", "r", encoding="utf-8") as fh:
  long_description = fh.read()
setuptools.setup(
    name='ConversationAgent',
    version='0.3.6',
    description='ConversationAgent',
    author='Theta',
    license='',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    package_data={
        "": ["*.txt", "*.py", "*.md"]
    },
    zip_safe=False
)




from ..LibStage import gen_multi_agent, QAStage, gen_agent
from .. import mock_client_once
import json

"""
# 預設腳本 迴圈式腳本
"""
bot_json = {
    "__MAIN_STAGES__": [
        {
            "stage_type": "__QA_STAGE__",
            "__STAGE_NAME__": "__開始階段__",
            "question": {
                "sys_welcome": "",
                "sys_refuse": "初始句拒絕",
                "sys_complete": ""
            },
            "qa_threshold": 0.8,
            "corpus": {
                "問答服務": "問答服務",
                "商品推薦": "商品推薦",
                "哈囉": "打招呼",
            },
            "__SAVED_NAME__": {
                "__QA_RESPOND__": "hello_r1",
                "__QA_RESPOND_THRESHOLD__": "hello_th",
                "__QA_RESPOND_QUESTION__": "hello_q1",
                "__QA_RESPOND_SCORE__": "hello_s1",
                "__RUNNING_CORPUS__": "hello_c1"
            },
            "__DISABLE_WELCOME__": True
        },
        {
            "stage_type": "__LIB_SWITCH_STAGE__",
            "__STAGE_NAME__": "__開始 切換階段__",
            "stages_filter": [
                [
                    "hello_r1",
                    "問答服務",
                    "_FAQ_"
                ],
                [
                    "hello_r1",
                    "商品推薦",
                    "_商品推薦_"
                ],
                [
                    "hello_r1",
                    "打招呼",
                    "_打招呼_"
                ]
            ]
        }
    ],
    "_打招呼_": [
        {
            "stage_type": "__RE_STAGE__",
            "__STAGE_NAME__": "__打招呼階段__",
            "question": {
                "sys_welcome": "",
                "sys_refuse": "",
                "sys_complete": "打招呼完成"
            },
            "__DISABLE_WELCOME__": True,
            "__DISABLE_REFUSE__": True
        }
    ],
    "_商品推薦_": [
        {
            "stage_type": "__RE_STAGE__",
            "__STAGE_NAME__": "__商品推薦階段__",
            "question": {
                "sys_welcome": "",
                "sys_refuse": "",
                "sys_complete": "推薦完成"
            },
            "__DISABLE_WELCOME__": True
        }
    ],
    "_FAQ_": [
        {
            "stage_type": "__QA_STAGE__",
            "__STAGE_NAME__": "__FAQ階段__",
            "question": {
                "sys_welcome": "FAQ歡迎",
                "sys_refuse": "FAQ拒絕",
                "sys_complete": ""
            },
            "qa_threshold": 0.7,
            "corpus": {
                "餐飲問題": "1",
                "營業時間": "2",
                "推薦商品": "3"
            },
            "__SAVED_NAME__": {
                "__QA_RESPOND__": "FAQ_r1",
                "__QA_RESPOND_THRESHOLD__": "FAQ_th",
                "__QA_RESPOND_QUESTION__": "FAQ_q1",
                "__QA_RESPOND_SCORE__": "FAQ_s1",
                "__RUNNING_CORPUS__": "FAQ_c1"
            }
        },
        {
            "stage_type": "__LIB_SWITCH_STAGE__",
            "__STAGE_NAME__": "__FAQ切換階段__",
            "stages_filter": [
                [
                    "FAQ_r1",
                    "1",
                    "_餐飲問題_"
                ],
                [
                    "FAQ_r1",
                    "2",
                    "_店舖問題_"
                ],
                [
                    "FAQ_r1",
                    "3",
                    "_商品推薦_"
                ]
            ]
        }
    ],
    "_循環FAQ_": [
        {
            "stage_type": "__QA_STAGE__",
            "__STAGE_NAME__": "__循環問答階段__",
            "question": {
                "sys_welcome": "循環FAQ歡迎",
                "sys_refuse": "循環FAQ拒絕",
                "sys_complete": ""
            },
            "qa_threshold": 0.7,
            "corpus": {
                "餐飲問題": "1",
                "營業時間": "2",
                "推薦商品": "3",
                "謝謝": "4"
            },
            "__SAVED_NAME__": {
                "__QA_RESPOND__": "FAQ2_r1",
                "__QA_RESPOND_THRESHOLD__": "FAQ2_th",
                "__QA_RESPOND_QUESTION__": "FAQ2_q1",
                "__QA_RESPOND_SCORE__": "FAQ2_s1",
                "__RUNNING_CORPUS__": "FAQ2_c1"
            }
        },
        {
            "stage_type": "__LIB_SWITCH_STAGE__",
            "__STAGE_NAME__": "__循環問答切換階段__",
            "stages_filter": [
                [
                    "FAQ2_r1",
                    "1",
                    "_餐飲問題_"
                ],
                [
                    "FAQ2_r1",
                    "2",
                    "_店舖問題_"
                ],
                [
                    "FAQ2_r1",
                    "3",
                    "_商品推薦_"
                ],
                [
                    "FAQ2_r1",
                    "4",
                    "_結束_"
                ]
            ]
        }
    ],
    "_店舖問題_": [
        {
            "stage_type": "__RE_STAGE__",
            "__STAGE_NAME__": "__店舖問題階段__",
            "question": {
                "sys_complete": "店舖問題 完成",
                "sys_refuse": "",
                "sys_welcome": ""
            },
            "__DISABLE_WELCOME__": True
        },
        {
            "stage_type": "__LIB_SWITCH_STAGE__",
            "stages_filter": [
                [
                    "*",
                    True,
                    "_循環FAQ_"
                ]
            ]
        }
    ],
    "_餐飲問題_": [
        {
            "stage_type": "__RE_STAGE__",
            "__STAGE_NAME__": "__餐飲問題階段__",
            "question": {
                "sys_complete": "飲品問題 完成",
                "sys_refuse": "",
                "sys_welcome": ""
            },
            "__DISABLE_WELCOME__": True
        },
        {
            "stage_type": "__LIB_SWITCH_STAGE__",
            "__STAGE_NAME__": "__餐飲問題切換階段__",
            "stages_filter": [
                [
                    "*",
                    True,
                    "_循環FAQ_"
                ]
            ]
        }

    ],
    "_結束_": [
        {
            "stage_type": "__RE_STAGE__",
            "__STAGE_NAME__": "__結束階段__",
            "question": {
                "sys_welcome": "",
                "sys_refuse": "",
                "sys_complete": "_結束_ 結束"
            },
            "__DISABLE_WELCOME__": True
        }
    ]
}


def output2file(data, filename="sample.json"):
    with open(f"./{filename}", "w") as f:
        json.dump(data, f, ensure_ascii=False)


data = {}
agent = gen_agent(bot_json)

#
text = "hi"
reply_text, data = mock_client_once(agent, text, data)
print(f"reply_text: {reply_text}")
# print(f"data: {data}")
assert ['初始句拒絕'] == reply_text

#
text = "問答服務"
reply_text, data = mock_client_once(agent, text, data)
print(f"reply_text: {reply_text}")
# print(f"data: {data}")
assert ['FAQ歡迎'] == reply_text

#
text = "餐飲問題"
reply_text, data = mock_client_once(agent, text, data)
print(f"reply_text: {reply_text}")
# print(f"data: {data}")
# output2file(data, "0.json")

assert ['飲品問題完成', '循環FAQ歡迎'] == reply_text

#
text = "餐飲問題"
reply_text, data = mock_client_once(agent, text, data)
print(f"reply_text: {reply_text}")
# print(f"data: {data}")
# output2file(data, "1.json")

assert ['飲品問題完成', '循環FAQ歡迎'] == reply_text

#
text = "餐飲問題"
reply_text, data = mock_client_once(agent, text, data)
print(f"reply_text: {reply_text}")
# print(f"data: {json.dumps(data,ensure_ascii=False)}")
# output2file(data, "2.json")
assert ['飲品問題完成', '循環FAQ歡迎'] == reply_text


#
text = "餐飲問題"
reply_text, data = mock_client_once(agent, text, data)
print(f"reply_text: {reply_text}")
# print(f"data: {data}")
# output2file(data, "3.json")
assert ['飲品問題完成', '循環FAQ歡迎'] == reply_text

#
text = "餐飲問題"
reply_text, data = mock_client_once(agent, text, data)
print(f"reply_text: {reply_text}")
# print(f"data: {data}")
# output2file(data, "4.json")
assert ['飲品問題完成', '循環FAQ歡迎'] == reply_text



from ..LibStage import gen_multi_agent, QAStage, gen_agent
from .. import mock_client_once

"""
# 預設腳本
"""
bot_json = {
    "__MAIN_STAGES__": [
        {
            "stage_type": "__RE_STAGE__",
            "__SYS_QUESTION__": {
                "__SYS_WELCOME__": "歡迎句",
                "__SYS_REFUSE__": "拒絕句",
                "__SYS_COMPLETE__": "完成句"
            },
            "is_fits": [
                [
                    "(通過條件)+",
                    "set_level"
                ]
            ]
        }
    ]
        
}
data = {}
agent = gen_agent(bot_json)

#
text = "hi"
reply_text, data = mock_client_once(agent, text, data)
print(f"data: {data}")
assert ['歡迎句'] == reply_text

text = "不要通過"
reply_text, data = mock_client_once(agent, text, data)
print(f"data: {data}")
assert ['拒絕句'] == reply_text

text = "通過條件"
reply_text, data = mock_client_once(agent, text, data)
print(f"data: {data}")
assert ['完成句'] == reply_text


"""
# 關閉Welcome問題
"""
bot_json = {
    "__MAIN_STAGES__": [
        {
            "stage_type": "__RE_STAGE__",
            "question": {
                "sys_reply_q1": "歡迎句",
                "sys_reply_q2": "拒絕句",
                "sys_reply_complete": "完成句"
            },
            "is_fits": [
                [
                    "(通過條件)+",
                    "set_level"
                ]
            ],
            "__DISABLE_WELCOME__": True
        }
    ]
        
}
data = {}
agent = gen_multi_agent(bot_json)

text = "不要通過"
reply_text, data = mock_client_once(agent, text, data)
print(f"data: {data}")
assert ['拒絕句'] == reply_text, f"['拒絕句'] == {reply_text}"

text = "通過條件"
reply_text, data = mock_client_once(agent, text, data)
print(f"data: {data}")
assert ['完成句'] == reply_text, f"['完成句'] == {reply_text}"

"""
# 關閉REFUSE問題
"""

bot_json = {
    "__MAIN_STAGES__": [
        {
            "stage_type": "__RE_STAGE__",
            "question": {
                "sys_reply_q1": "歡迎句",
                "sys_reply_q2": "拒絕句",
                "sys_reply_complete": "完成句"
            },
            "is_fits": [
                [
                    "(通過條件)+",
                    "set_level"
                ]
            ],
            "__DISABLE_REFUSE__": True
        }
    ]

}
data = {}
agent = gen_multi_agent(bot_json)

#
text = "hi"
reply_text, data = mock_client_once(agent, text, data)
print(f"data: {data}")
assert ['歡迎句'] == reply_text, f"['歡迎句'] == {reply_text}"


text = "測試"
reply_text, data = mock_client_once(agent, text, data)
print(f"data: {data}")
assert ['完成句'] == reply_text, f"['完成句'] == {reply_text}"
